from typing import Literal
from dataclasses import dataclass
from enum import Enum
from typing import Iterable, Generator
from PIL.Image import Image
from fitz import Document
from doc_page_extractor import clip, PaddleLang, Rectangle, Layout, LayoutClass, OCRFragment, ExtractedResult
from .document import DocumentExtractor


class TextKind(Enum):
  TITLE = 0
  PLAIN_TEXT = 1
  ABANDON = 2

@dataclass
class Text:
  content: str
  rank: float
  rect: Rectangle

@dataclass
class TextBlock:
  rect: Rectangle
  kind: TextKind
  texts: list[Text]
  has_paragraph_indentation: bool = False
  last_line_touch_end: bool = False

@dataclass
class AssetBlock:
  rect: Rectangle
  image: Image
  texts: list[Text]

Block = TextBlock | AssetBlock

class PDFPageExtractor:
  def __init__(
      self,
      device: Literal["cpu", "cuda"],
      model_dir_path: str,
      debug_dir_path: str | None = None,
    ):
    self._doc_extractor: DocumentExtractor = DocumentExtractor(
      device=device,
      model_dir_path=model_dir_path,
      debug_dir_path=debug_dir_path,
    )

  def extract(self, pdf: str | Document, lang: PaddleLang) -> Generator[list[Block], None, None]:
    for result, layouts in self._doc_extractor.extract(pdf, lang):
      blocks = self._convert_to_blocks(result, layouts)
      page_range = self._texts_range(blocks)

      for block in blocks:
        if not isinstance(block, TextBlock) or \
           block.kind == TextKind.ABANDON:
          continue

        if len(block.texts) == 1:
          mean_line_height, x1, x2 = page_range
        else:
          mean_line_height, x1, x2 = self._texts_range((block,))

        first_text = block.texts[0]
        last_text = block.texts[-1]
        first_delta_x = (first_text.rect.lt[0] + first_text.rect.lb[0]) / 2 - x1
        last_delta_x = x2 - (last_text.rect.rt[0] + last_text.rect.rb[0]) / 2
        block.has_paragraph_indentation = first_delta_x > mean_line_height
        block.last_line_touch_end = last_delta_x < mean_line_height

      yield blocks

  def _convert_to_blocks(self, result: ExtractedResult, layouts: list[Layout]) -> list[Block]:
    store: list[tuple[LayoutClass, Block]] = []
    def previous_block(cls: LayoutClass) -> Block | None:
      for i in range(len(store) - 1, -1, -1):
        block_cls, block = store[i]
        if cls == block_cls:
          return block
        if cls != LayoutClass.ABANDON:
          return None
      return None

    for layout in layouts:
      cls = layout.cls
      if cls == LayoutClass.TITLE:
        store.append((cls, TextBlock(
          rect=layout.rect,
          kind=TextKind.TITLE,
          texts=self._convert_to_text(layout.fragments),
        )))
      elif cls == LayoutClass.PLAIN_TEXT:
        store.append((cls, TextBlock(
          rect=layout.rect,
          kind=TextKind.PLAIN_TEXT,
          texts=self._convert_to_text(layout.fragments),
        )))
      elif cls == LayoutClass.ABANDON:
        store.append((cls, TextBlock(
          rect=layout.rect,
          kind=TextKind.ABANDON,
          texts=self._convert_to_text(layout.fragments),
        )))
      elif cls == LayoutClass.FIGURE or \
           cls == LayoutClass.TABLE or \
           cls == LayoutClass.ISOLATE_FORMULA:
        store.append((cls, AssetBlock(
          rect=layout.rect,
          texts=[],
          image=clip(result, layout),
        )))
      elif cls == LayoutClass.FIGURE_CAPTION:
        block = previous_block(LayoutClass.FIGURE)
        if block is not None:
          assert isinstance(block, AssetBlock)
          block.texts.extend(self._convert_to_text(layout.fragments))
      elif cls == LayoutClass.TABLE_CAPTION or \
           cls == LayoutClass.TABLE_FOOTNOTE:
        block = previous_block(LayoutClass.TABLE)
        if block is not None:
          assert isinstance(block, AssetBlock)
          block.texts.extend(self._convert_to_text(layout.fragments))
      elif cls == LayoutClass.FORMULA_CAPTION:
        block = previous_block(LayoutClass.ISOLATE_FORMULA)
        if block is not None:
          assert isinstance(block, AssetBlock)
          block.texts.extend(self._convert_to_text(layout.fragments))

    return [block for _, block in store]

  def _texts_range(self, blocks: Iterable[Block]) -> tuple[float, float, float]:
    sum_lines_height: float = 0.0
    texts_count: int = 0
    x1: float = float("inf")
    x2: float = float("-inf")

    for block in blocks:
      if not isinstance(block, TextBlock):
        continue
      if block.kind == TextKind.ABANDON:
        continue
      for text in block.texts:
        sum_lines_height += text.rect.size[1]
        texts_count += 1
        for x, _ in text.rect:
          x1 = min(x1, x)
          x2 = max(x2, x)

    if texts_count == 0:
      return 0.0, 0.0, 0.0
    return sum_lines_height / texts_count, x1, x2

  def _convert_to_text(self, fragments: list[OCRFragment]) -> list[Text]:
    return [
      Text(
        content=f.text,
        rank=f.rank,
        rect=f.rect,
      )
      for f in fragments
    ]


import unittest

from typing import Iterable
from doc_page_extractor import Rectangle, Layout, LayoutClass, OCRFragment
from pdf_craft.pdf.section import Section
from pdf_craft.pdf.text_matcher import check_texts_matching_rate, split_into_words

_Rect = tuple[float, float, float, float]

class TextFramework(unittest.TestCase):

  def test_framework_generation(self):
    layouts1 = [
      _layout((20.0, 20.0, 100.0, 20.0), (
        ((1.5, 2.3, 197.0, 11.0), "the Little Prince"),
      )),
      _layout((160.0, 30.0, 30.0, 40.0), (
        ((5.0, 5.0, 20.0, 15.0), "Antoine de"),
        ((4.0, 21.0, 22.0, 14.8), "Saint-Exupéry"),
      )),
      _layout((23.0, 50.0, 130.0, 50.0), (
        ((0.0, 0.0, 135.0, 15.0), "I believe that for his escape he took advantage of"),
        ((0.0, 17.0, 135.0, 15.0), "the migration of a flock of wild birds. On the morning"),
        ((0.0, 34.0, 135.0, 15.0), "of his departure he put his planet in perfect order. "),
      )),
      _layout((23.0, 120.0, 170.0, 230.0), (
        ((0.0, 0.0, 165.0, 20.0), "He carefully cleaned out his active volcanoes. He possessed"),
        ((0.0, 25.0, 165.0, 20.0), "two active volcanoes; and they were very convenient for"),
        ((0.0, 50.0, 165.0, 20.0), "heating his breakfast in the morning. He also had one"),
        ((0.0, 75.0, 165.0, 20.0), "volcano that was extinct. But, as he said, \"One never knows!\","),
        ((0.0, 100.0, 165.0, 20.0), "So he cleaned out the extinct volcano, too. If they are well cleaned out,"),
        ((0.0, 125.0, 165.0, 20.0), "volcanoes burn slowly and steadily, without any eruptions."),
        ((0.0, 150.0, 165.0, 20.0), "Volcanic eruptions are like fires in a chimney."),
      )),
      _layout((90.0, 370.0, 40.0, 20.0), (
        ((1.0, 2.0, 35.0, 16.5), "36"),
      )),
    ]
    layouts2 = [
      _layout((160.3, 30.7, 30.2, 40.3), (
        ((5.0, 5.0, 20.0, 15.0), "Antoine  de"),
        ((4.0, 21.0, 22.0, 14.8), "Saint-Exupéry"),
      )),
      _layout((20.0, 20.0, 100.0, 20.0), (
        ((1.7, 2.2, 200.0, 11.2), "the Little Prince"),
      )),
      _layout((90.2, 370.2, 40.3, 20.2), (
        ((1.0, 2.0, 35.0, 16.5), "37"),
      )),
      _layout((23.7, 50.2, 130.3, 50.1), (
        ((0.0, 0.0, 135.0, 15.0), "The first of them was inhabited by a king."),
        ((0.0, 17.0, 135.0, 15.0), "Clad in royal purple and ermine, he was seated"),
        ((0.0, 34.0, 135.0, 15.0), "upon a throne which was at the same time both simple and majestic."),
      )),
    ]
    # Scanned book pages often shift, this is normal and needs to be fixed.
    delta2 = (27.35, 48.12)
    layouts2 = [
      Layout(
        cls=layout.cls,
        rect=_move_rect(layout.rect, delta2),
        fragments=[
          OCRFragment(
            order=fragment.order,
            text=fragment.text,
            rank=fragment.rank,
            rect=_move_rect(fragment.rect, delta2),
          )
          for fragment in layout.fragments
        ],
      )
      for layout in layouts2
    ]
    session = Section(layouts1)
    session.link_next(Section(layouts2), 1)
    framework_indexes = sorted([layouts1.index(layout) for layout in session.framework()])

    self.assertListEqual(framework_indexes, [0, 1, 4])

  def test_texts_matching(self):
    self.assertEqual(
      check_texts_matching_rate("Hello, world!", "Hello, world!"),
      (1.0, 4),
    )
    self.assertEqual(
      check_texts_matching_rate("围点打援", "围点打援！"),
      (4 / 5, 5),
    )
    self.assertEqual(
      check_texts_matching_rate("围点打援", "围点打缓"),
      (3 / 4, 4),
    )
    self.assertEqual(
      check_texts_matching_rate("围点打援", "援点打围"),
      (0.75, 4),
    )
    self.assertEqual(
      check_texts_matching_rate("围点Foobar打援", "围点打援"),
      (4 / 5, 5),
    )

  def test_splitting_into_words(self):
    self.assertEqual(
      list(split_into_words("Hello, world!")),
      ["Hello", ",", "world", "!"]
    )
    self.assertEqual(
      list(split_into_words("Люди мира едины.")),
      ["Люди", "мира", "едины", "."],
    )
    self.assertEqual(
      list(split_into_words("各个国家都有各个国家的国歌。")),
      [c for c in "各个国家都有各个国家的国歌。"]
    )
    self.assertEqual(
      list(split_into_words("获取class 的意义")),
      ["获", "取", "class", "的", "意", "义"]
    )

def _layout(rect: _Rect, fragments: Iterable[tuple[_Rect, str]]):
  origin = (rect[0], rect[1])
  return Layout(
    cls=LayoutClass.PLAIN_TEXT,
    rect=_rect(rect),
    fragments=[
      OCRFragment(
        order=i,
        text=text,
        rank=1.0,
        rect=_move_rect(_rect(rect), origin),
      )
      for i, (rect, text) in enumerate(fragments)
    ],
  )

def _rect(rect: _Rect) -> Rectangle:
  x0, y0, width, height = rect
  return Rectangle(
    lt=(x0, y0),
    lb=(x0, y0 + height),
    rt=(x0 + width, y0),
    rb=(x0 + width, y0 + height),
  )

def _move_rect(rect: Rectangle, delta: tuple[float, float]) -> Rectangle:
  return Rectangle(
    lt=(rect.lt[0] + delta[0], rect.lt[1] + delta[1]),
    lb=(rect.lb[0] + delta[0], rect.lb[1] + delta[1]),
    rt=(rect.rt[0] + delta[0], rect.rt[1] + delta[1]),
    rb=(rect.rb[0] + delta[0], rect.rb[1] + delta[1]),
  )


import io
import os

from hashlib import sha256
from typing import Iterable
from .pdf import Text, TextKind, TextBlock, AssetBlock, Block

class MarkDownWriter:
  def __init__(self, md_path: str, assets_path: str, encoding: str | None):
    self._assets_path: str = assets_path
    self._abs_assets_path: str = os.path.abspath(os.path.join(md_path, "..", assets_path))
    self._file: io.TextIOWrapper = open(md_path, "w", encoding=encoding)
    self._texts_buffer: list[Text] = []

  def __enter__(self) -> "MarkDownWriter":
    return self

  def __exit__(self, exc_type, exc_val, exc_tb) -> None:
    self.close()

  def flush(self) -> None:
    self._file.flush()

  def close(self) -> None:
    self._close_texts_buffer()
    self._file.close()

  def write(self, block: Block) -> None:
    if isinstance(block, TextBlock):
      if block.kind == TextKind.TITLE:
        self._close_texts_buffer()
        self._file.write("# ")
        self._write_text_contents(block.texts)
        self._file.write("\n\n")
      elif block.kind == TextKind.PLAIN_TEXT:
        self._write_plain_text(block)
    elif isinstance(block, AssetBlock):
      self._close_texts_buffer()
      self._write_image(block)

  def _close_texts_buffer(self):
    if len(self._texts_buffer) > 0:
      self._write_text_contents(self._texts_buffer)
      self._file.write("\n\n")
      self._texts_buffer.clear()

  def _write_plain_text(self, block: TextBlock) -> None:
    if block.has_paragraph_indentation:
      self._close_texts_buffer()
    self._texts_buffer.extend(block.texts)
    if not block.last_line_touch_end:
      self._close_texts_buffer()

  def _write_image(self, block: AssetBlock) -> None:
    os.makedirs(self._abs_assets_path, exist_ok=True)
    hash = sha256()
    hash.update(block.image.tobytes())
    file_name = f"{hash.hexdigest()}.png"
    file_path = os.path.join(self._abs_assets_path, file_name)
    relative_path = os.path.join(self._assets_path, file_name)

    if not os.path.exists(file_path):
      block.image.save(file_path, "PNG")

    self._file.write("![")
    self._write_text_contents(block.texts, "]")
    self._file.write(f"]({relative_path})")
    self._file.write("\n\n")

  def _write_text_contents(self, texts: Iterable[Text], ban_symbol: str | None = None) -> None:
    for text in texts:
      content = text.content.strip()
      content = content.replace("\n", " ")
      if ban_symbol is not None:
        content.replace(ban_symbol, "")
      self._file.write(content)


import os
import fitz

from typing import Generator, Literal
from PIL.Image import frombytes, Image
from doc_page_extractor import plot, Layout, PaddleLang, DocExtractor, ExtractedResult
from .section import Section


class DocumentExtractor:
  def __init__(
      self,
      device: Literal["cpu", "cuda"],
      model_dir_path: str,
      debug_dir_path: str | None = None,
    ):
    self._debug_dir_path: str | None = debug_dir_path
    self._doc_extractor = DocExtractor(
      device=device,
      model_dir_path=model_dir_path,
      order_by_layoutreader=False,
    )

  def extract(self, pdf: str | fitz.Document, lang: PaddleLang) -> Generator[tuple[ExtractedResult, list[Layout]], None, None]:
    for result, section in self._extract_results_and_sections(pdf, lang):
      framework_layouts = section.framework()
      yield result, [
        layout for layout in result.layouts
        if layout not in framework_layouts
      ]

  def _extract_results_and_sections(self, pdf: str | fitz.Document, lang: PaddleLang):
    max_len = 2 # section can be viewed up to 2 pages back
    queue: list[tuple[ExtractedResult, Section]] = []

    for result in self._extract_page_result(pdf, lang):
      section = Section(result.layouts)
      for i, (_, pre_section) in enumerate(queue):
        offset = len(queue) - i
        pre_section.link_next(section, offset)

      queue.append((result, section))
      if len(queue) > max_len:
        yield queue.pop(0)

    for result, section in queue:
      yield result, section

  def _extract_page_result(self, pdf: str | fitz.Document, lang: PaddleLang):
    if self._debug_dir_path is not None:
      os.makedirs(self._debug_dir_path, exist_ok=True)

    document: fitz.Document
    should_close = False
    if isinstance(pdf, str):
      document = fitz.open(pdf)
      should_close = True
    else:
      document = pdf

    try:
      for i, page in enumerate(document.pages()):
        dpi = 300 # for scanned book pages
        image = self._page_screenshot_image(page, dpi)
        result = self._doc_extractor.extract(
          image=image,
          lang=lang,
          adjust_points=False,
        )
        if self._debug_dir_path is not None:
          self._generate_plot(image, i, result, self._debug_dir_path)
        yield result
    finally:
      if should_close:
        document.close()

  def _page_screenshot_image(self, page: fitz.Page, dpi: int):
    default_dpi = 72
    matrix = fitz.Matrix(dpi / default_dpi, dpi / default_dpi)
    pixmap = page.get_pixmap(matrix=matrix)
    return frombytes("RGB", (pixmap.width, pixmap.height), pixmap.samples)

  def _generate_plot(self, image: Image, index: int, result: ExtractedResult, plot_path: str):
    plot_image: Image
    if result.adjusted_image is None:
      plot_image = image.copy()
    else:
      plot_image = result.adjusted_image

    plot(plot_image, result.layouts)
    os.makedirs(plot_path, exist_ok=True)
    image_path = os.path.join(plot_path, f"plot_{index + 1}.png")
    plot_image.save(image_path)


from setuptools import setup, find_packages

setup(
  name="pdf-craft",
  version="0.0.1",
  author="Tao Zeyu",
  author_email="i@taozeyu.com",
  url="https://github.com/oomol-lab/pdf-craft",
  description="PDF craft can convert PDF files into various other formats. This project will focus on processing PDF files of scanned books. The project has just started.",
  packages=find_packages(),
  long_description=open("./README.md", encoding="utf8").read(),
  long_description_content_type="text/markdown",
  install_requires=[
    "pyMuPDF>=1.25.3,<2.0.0",
    "alphabet-detector>=0.0.7,<1.0.0",
    "shapely>=2.0.6,<3.0.0",
    "doc-page-extractor==0.0.1",
  ],
)